﻿using UnityEngine;
using System.Collections.Generic;
using MeshVR;
using MeshedVR;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using System.Linq;
using MVR.FileManagementSecure;

/* /////////////////////////////////////////////////////////////////////////////////////////////////

Author: PluginIdea 
Email:819661664@qq.com
Patreon:https://www.patreon.com/PluginIdea

©2022 PluginIdea, Inc. All rights reserved.

///////////////////////////////////////////////////////////////////////////////////////////////// */

namespace PluginIdea
{

    public class SlideControl : MVRScript
    {

        public const string InputModule = "InputModule";

        private List<string> playModes = new List<string> { "Random", "Sequence", "Sequence Inverse" };
        private List<string> transitionTypes = new List<string> { "Slide_Shutter5"};
        private List<string> aspectRadioPresets = new List<string> { "Custom", "16:9", "16:10", "4:3", "2:1", "1:1" };
        private Dictionary<string, int> dic_EaseType = new Dictionary<string, int> {
            { "easeInQuad", iTween.EaseType.easeInQuad },
            { "easeOutQuad", iTween.EaseType.easeOutQuad },
            { "linear", iTween.EaseType.linear },
        };
        private List<string> imageList = new List<string>();
        private List<Texture2D> cachedImages = new List<Texture2D>();
        private List<int> randomPool = new List<int>();

        private JSONStorableStringChooser playModeChooser;
        private JSONStorableStringChooser aspectRadioPresetChooser;
        private JSONStorableStringChooser transitionAnimChooser;
        private JSONStorableStringChooser transitionTypeChooser;
        private JSONStorableBool isAutoPlay;
        private JSONStorableBool isRandomTransitionType;
        private JSONStorableFloat switchingInterval;
        private JSONStorableUrl folderPath;
        private JSONStorableFloat width, height;
        private JSONStorableFloat translateDuration;
        private JSONStorableBool isPreloadTextures;
        private JSONStorableFloat preloadAmount;
        private JSONStorableBool islockWhRadio;

        public JSONStorableAction nextImgAction;
        public JSONStorableAction prevImgAction;

        private Dictionary<string, SlideModuleBase> uid2Modules = new Dictionary<string, SlideModuleBase>();
        private Dictionary<string, Shader> dic_BundleShaders = new Dictionary<string, Shader>();

        private float timer;
        private bool isExternalResourcesLoaded;
        private float whAspectValue = 1;

        private string lastTransitionType;

        private string currentFolderUrl;

        private bool IsInitied
        {
            get
            {
                return isExternalResourcesLoaded;
            }
        }

        public Dictionary<string, Shader> Dic_BundleShaders
        {
            get
            {
                return this.dic_BundleShaders;
            }
        }

        private GameObject bgPlane;
        private GameObject rdPlane;
        private Renderer rd, bgRd;
        private Renderer DisplayRender
        {
            get
            {
                if (rd == null)
                {
                    rd = rdPlane.GetComponent<Renderer>();
                }
                return rd;
            }
        }

        public override void Init()
        {
            base.Init();

            if (containingAtom.type != "Empty")
            {
                SuperController.LogError("need add as Empty plugin");
                return;
            }

            ResourceInit();
            DataInit();
            UIInit();
            EventInit();
            LogicInit();
        }

        private void ResourceInit()
        {
            string assetsBundlePath = Utils.GetPluginPath(this) + "/resources/slidecontrolfree.assetbundle";
            Request request = new Request { path = assetsBundlePath, callback = OnAssetBundleLoaded };
            AssetLoader.QueueLoadAssetBundleFromFile(request);
        }

        private void OnAssetBundleLoaded(Request request)
        {
            Shader[] shaders = request.assetBundle.LoadAllAssets<Shader>();
            for (int i = 0; i < shaders.Length; i++)
            {
                SuperController.LogMessage("Shader: " + shaders[i].name);
                if (!dic_BundleShaders.ContainsKey(shaders[i].name))
                {
                    string key = shaders[i].name;
                    dic_BundleShaders.Add(key, shaders[i]);
                }
            }

            isExternalResourcesLoaded = true;

            bgRd.material.shader = dic_BundleShaders["PluginIdea/Slide_bgPlane"];

            // 重新load
            folderPath.Reload();
            width.setCallbackFunction(width.val);
            transitionTypeChooser.setCallbackFunction(transitionTypeChooser.val);
        }

        #region Init Func
        private void DataInit()
        {
        }

        private GameObject loadingIcon;
        private void UIInit()
        {
            loadingIcon = SuperController.singleton.loadingIcon.Find("LoadingIconCanvas/LoadingIcon").gameObject;
            loadingIcon.SetActive(false);

            Utils.OnInitUI(CreateUIElement);

            Utils.SetupInfoText(this, "<size=40><b>SlideControl(Free) 1.0</b></size>\n\n" +
                "This is a Simple Slide Player for Texture.\n\n" +
                 "<b>Is Auto Play:</b>Whether to play automatically. Otherwise, you need to switch the next one manually.\n\n" +
                 "<b>Set Folder:</b>Due to the limitations of the game, you actually need to choose an actual picture here.\n\n" +
                 "<b>TransitionType:</b>Types of transition special effects.\n\n" +
                 "<b>AnimationEaseType:</b>Transition animation curve."
                 , 500, true);

            //Utils.SetupInfoTextNoScroll(this, "<size=40><b>Common Parameters</b></size>", 45, false);
            Utils.SetupInfoTextNoScroll(this, "<size=30><b>Transition Effect</b></size>", 35, true);

            transitionAnimChooser = Utils.SetupStringChooser(this, "AnimationEaseType", dic_EaseType.Keys.ToList(), 2, true);

            playModeChooser = Utils.SetupStringChooser(this, "Playback Order", playModes, 1, false);

            isAutoPlay = Utils.SetupToggle(this, "Is Auto Play", true, false);

            Utils.SetupTwinButton(this, "Prev Image", () => {
                prevImgAction.actionCallback();
            }, "Next Image", () => {
                nextImgAction.actionCallback();
            }, false);

            folderPath = Utils.SetupFolderChooser(this, "Set Folder", $"Custom/Assets/MyAssets", false, null);
            folderPath.setCallbackFunction = (url) => {
                if (!isExternalResourcesLoaded) return;
                currentFolderUrl = url;
                //SuperController.LogMessage(url);
                string dir = FileManagerSecure.GetDirectoryName(url);
                string[] files = FileManagerSecure.GetFiles(dir);
                if (files.Length > 0)
                {
                    if (currentTex != null) Destroy(currentTex);
                    if (nextTex != null) Destroy(nextTex);
                    DestroyCachedImages();

                    imageList.Clear();
                    files.ToList().ForEach(p => {
                        if (p.EndsWith(".jpg") || p.EndsWith(".png") || p.EndsWith(".jpeg") || p.EndsWith(".tiff"))
                        {
                            imageList.Add(p);
                        }
                    });

                    if (imageList.Count <= 1)
                    {
                        SuperController.LogError("The folder needs to contain at least two valid pictures!");
                        return;
                    }

                    // preload
                    if (isPreloadTextures.val)
                    {
                        int tmpIndex = 0;
                        imageList.ForEach(p => {
                            Utils.LoadTexture(p, new TextureSettings(), (tex) => {
                                if (tex != null)
                                {
                                    cachedImages.Add(tex);

                                    // set Begin Texture
                                    if (tmpIndex == 0)
                                    {
                                        currentTex = tex;
                                        isCurTexLoadFinished = true;
                                    }

                                    if (tmpIndex == 1)
                                    {
                                        nextTex = tex;
                                        isNextTexLoadFinished = true;
                                        OnTexLoadFinished();
                                    }

                                    tmpIndex++;
                                }
                            });
                        });
                    }
                    else
                    {
                        Utils.LoadTexture(imageList[imgIndex], new TextureSettings(), (tex) => {
                            currentTex = tex;
                            isCurTexLoadFinished = true;
                            OnTexLoadFinished();
                        });

                        Utils.LoadTexture(imageList[imgIndex + 1], new TextureSettings(), (tex) => {
                            nextTex = tex;
                            isNextTexLoadFinished = true;
                            OnTexLoadFinished();
                        });
                    }

                    imgIndex = 1;
                }
            };

            Utils.SetupInfoTextNoScroll(this, "<size=30><b>Parameter</b></size>", 35, false);
            switchingInterval = Utils.SetupSliderFloat(this, "Switch Interval", 5f, 0.5f, 100, false);

            islockWhRadio = Utils.SetupToggle(this, "IsLockWidthHeightRadio", false, false);

            isRandomTransitionType = Utils.SetupToggle(this, "IsRandomTransitionType", false, true);

            transitionTypeChooser = Utils.SetupStringChooser(this, "TransitionType", transitionTypes, true);
            transitionTypeChooser.setCallbackFunction = (val) => {
                if (!isExternalResourcesLoaded) return;
                string key = string.Format("PluginIdea/{0}", val);
                DisplayRender.material.shader = dic_BundleShaders[key];
            };

            aspectRadioPresetChooser = Utils.SetupStringChooser(this, "AspectRadios", aspectRadioPresets, false);
            aspectRadioPresetChooser.setCallbackFunction = (val) => {
                switch (val)
                {
                    case "16:9": whAspectValue = 1.777f; break;
                    case "16:10": whAspectValue = 1.6f; break;
                    case "4:3": whAspectValue = 1.333f; break;
                    case "2:1": whAspectValue = 2f; break;
                    case "1:1": whAspectValue = 1f; break;
                    default:
                        break;
                }
            };

            translateDuration = Utils.SetupSliderFloat(this, "Transition Duration", 1.5f, 0.1f, 5f, true);

            width = Utils.SetupSliderFloat(this, "Width", 1.6f, 0.1f, 20f, false);
            width.setCallbackFunction = (val) => {
                if (islockWhRadio.val)
                {
                    float h = val / whAspectValue;
                    rdPlane.transform.localScale = new Vector3(val, h, 1);
                    height.val = h;
                }
                else
                {
                    rdPlane.transform.localScale = new Vector3(val, rdPlane.transform.localScale.y, 1);
                }
            };

            height = Utils.SetupSliderFloat(this, "Height", 1f, 0.1f, 20f, false);
            height.setCallbackFunction = (val) =>
            {
                if (islockWhRadio.val)
                {
                    float x = val * whAspectValue;
                    rdPlane.transform.localScale = new Vector3(x, val, 1);
                    width.val = x;
                }
                else
                {
                    rdPlane.transform.localScale = new Vector3(rdPlane.transform.localScale.x, val, 1);
                }
            };

            Utils.SetupInfoTextNoScroll(this, "<size=30><b>Extern</b></size>", 35, false);
            isPreloadTextures = Utils.SetupToggle(this, "Is Preload Textures", false, false);
        }

        private void EventInit()
        {
            RegisterModule(InputModule, new InputModule(this));

            nextImgAction = Utils.SetupAction(this, "NextImageAction", () =>
            {
                SetNextImgPlayIndex();
                SwitchImage();
            });

            prevImgAction = Utils.SetupAction(this, "PrevImageAction", () => {
                SetPrevImgPlayIndex();
                SwitchImage();
            });
        }

        private void LogicInit()
        {
            CreateQuadObject();
        }

        #endregion

        private Texture2D currentTex, nextTex;
        private bool isCurTexLoadFinished, isNextTexLoadFinished;
        int imgIndex = 0;
        private void SwitchImage()
        {
            if (nextTex == null)
            {
                int nextImgIndex = imgIndex + 1;
                if (nextImgIndex == imageList.Count) nextImgIndex = 0;
                if (isPreloadTextures.val)
                {
                    nextTex = cachedImages[nextImgIndex];
                    isNextTexLoadFinished = true;
                    OnTexLoadFinished();
                }
                else
                {
                    Utils.LoadTexture(imageList[nextImgIndex], new TextureSettings(), (tex) => {
                        nextTex = tex;
                        isNextTexLoadFinished = true;
                        OnTexLoadFinished();
                    });
                }
            }

            //SuperController.LogMessage(string.Format("index: {0}, imglistCount: {1}, cacheCount: {2}", imgIndex, imageList.Count, cachedImages.Count));

            // Next Round
            if (isAutoPlay.val)
            {
                switch (playModeChooser.val)
                {
                    case "Random":
                        SetRandomImgPlayIndex();
                        break;
                    case "Sequence":
                        SetNextImgPlayIndex();
                        break;
                    case "Sequence Inverse":
                        SetPrevImgPlayIndex();
                        break;
                    default:
                        break;
                }
            }
        }

        private void SetPrevImgPlayIndex()
        {
            imgIndex--;
            if (imgIndex < 0) imgIndex = imageList.Count - 1;
        }

        private void SetNextImgPlayIndex()
        {
            imgIndex++;
            if (imgIndex == imageList.Count) imgIndex = 0;
        }

        private void SetRandomImgPlayIndex()
        {
            if (randomPool.Count == 0)
            {
                for (int i = 0; i < imageList.Count; i++)
                {
                    randomPool.Add(i);
                }
            }

            int tmpindex = Random.Range(0, randomPool.Count);
            imgIndex = randomPool[tmpindex];
            randomPool.RemoveAt(tmpindex);
        }

        private void OnUpdate(float progress)
        {
            //SuperController.LogMessage(progress.ToString());
            DisplayRender.material.SetFloat("_Progress", progress);
        }

        private void OnTransitionAnimPlayFinished()
        {
            if (!isPreloadTextures.val)
                Destroy(currentTex);
            currentTex = nextTex;
            DisplayRender.material.SetTexture("_ForegroundTex", currentTex);
            DisplayRender.material.SetFloat("_Progress", 0);
            nextTex = null;
            isNextTexLoadFinished = false;
        }

        private void OnTexLoadFinished()
        {
            if (isCurTexLoadFinished && isNextTexLoadFinished)
            {
                //DisplayRender.material.SetTexture("_ForegroundTex", currentTex);
                bgRd.material.SetTexture("_BackgroundTex", nextTex);

                // random transitionType
                if (isRandomTransitionType.val) transitionTypeChooser.val = transitionTypes[Random.Range(0, transitionTypes.Count)];
                iTween.ValueTo(gameObject, iTween.Hash("from", 0, "to", 1, "easetype", dic_EaseType[transitionAnimChooser.val].ToString(), "onupdate", "OnUpdate", "oncomplete", "OnTransitionAnimPlayFinished", "time", translateDuration.val));
            }
        }

        private void CreateQuadObject()
        {
            rdPlane = GameObject.CreatePrimitive(PrimitiveType.Quad);
            rdPlane.GetComponent<Collider>().enabled = false;
            rdPlane.transform.SetParent(containingAtom.reParentObject.Find("object/rescaleObject"));
            rdPlane.transform.localScale = new Vector3(width.val, height.val, 1);
            rdPlane.transform.localPosition = Vector3.zero;
            rdPlane.transform.localRotation = Quaternion.identity;

            bgPlane = GameObject.CreatePrimitive(PrimitiveType.Quad);
            bgPlane.GetComponent<Collider>().enabled = false;
            bgPlane.transform.SetParent(rdPlane.transform);
            bgPlane.transform.localPosition = Vector3.zero;
            bgPlane.transform.localRotation = Quaternion.identity;
            bgPlane.transform.localScale = Vector3.one;
            bgPlane.transform.position += bgPlane.transform.forward * 0.001f;
            bgRd = bgPlane.GetComponent<Renderer>();

            //bgPlane.transform.SetParent(rdPlane.transform.parent);
            //bgPlane.SetActive(false);
        }

        private void Update()
        {
            // Update
            foreach (var module in uid2Modules.Values.ToList())
            {
                module.OnUpdate();
            }

            // Tick
            if (isAutoPlay.val && IsInitied && imageList.Count > 1)
            {
                timer += Time.deltaTime;
                if (timer > switchingInterval.val)
                {
                    timer = 0;
                    SwitchImage();
                }
            }
        }

        private void RegisterModule(string key, SlideModuleBase module)
        {
            if (!uid2Modules.ContainsKey(key))
            {
                uid2Modules.Add(key, module);
                module.Init();
            }
        }

        private void UnRegisterModule(string key)
        {
            if (uid2Modules.ContainsKey(key))
            {
                uid2Modules[key].UnInit();
                uid2Modules.Remove(key);
            }
        }

        private void OnPostRender()
        {
            foreach (var module in uid2Modules.Values.ToList())
            {
                module.OnPostRender();
            }
        }

        private void OnDestroy()
        {
            if (uid2Modules.Count > 0)
            {
                foreach (var key in uid2Modules.Keys.ToList())
                {
                    UnRegisterModule(key);
                }
                uid2Modules.Clear();
            }

            if (rdPlane != null)
            {
                rd = null;
                Destroy(rdPlane);
            }

            DestroyCachedImages();

            Utils.OnDestroyUI();

            loadingIcon.SetActive(true);
        }

        private void DestroyCachedImages()
        {
            if (cachedImages.Count > 0)
            {
                cachedImages.ForEach(img => {
                    if (img != null) Destroy(img);
                });
                cachedImages.Clear();
            }
        }
    }
}

